""" Functions in this module check how much time in each audio recording has been listened to for annotation purposes.
They start with the outputs produced by recap_regions_listen_time_compute.py in the `annot_distr` repository
(https://github.com/SeedlingsBabylab/annot_distr/). More specifically, here we use the region boundaries extracted
from each cha file and put into output/cha_structures/<child>_<month>_sparse_code.cha.txt that look approximately like
this:

    subregion starts 1200
    subregion ends 2400
    ...
    surplus starts 4500
    surplus end 5000

    Position: 1, Rank: 1
    Position: 2, Rank: 5
    Position: 3, Rank: 2
    Position: 4, Rank: 4
    Position: 5, Rank: 3

Recordings for months 6 and 7 do not have the subregion information and are already assessed in the annot_distr repo,
so we do not work with them here.

For each subregion, we calculate the amount of time listened to within that subregion accounting for skips and silences.
Starting from the subregion ranked the first on talkativeness, we then assign makeup and extra regions to each subregion
until the total listened time is not at least an hour.
"""
from enum import Enum
import re

import pandas as pd

# Two recordings have four subregions, all the other one have five
NO_CODEABLE_WORDS_BUT_LISTENED_COMMENT = 'subregion has been listened to but contains no codeable words'
DEFAULT_SUBREGION_COUNT = 5
RECORDINGS_WITH_FOUR_SUBREGIONS = ((21, 14), (45, 10))


class RegionType(Enum):
    SUBREGION = 'subregion'
    SILENCE = 'silence'
    SKIP = 'skip'
    MAKEUP = 'makeup'
    EXTRA = 'extra'
    SURPLUS = 'surplus'


# List of values to check with `in`
REGION_TYPES = [rt.value for rt in RegionType]

# Number of decimal places used when converting milliseconds to hours
PRECISION = 2

# This is a very permissive regex copied from annot_distr,
ANNOTATION_REGEX = re.compile(
    r'([a-zA-Z][a-z+]*)( +)(&=)([A-Za-z]{1})(_)([A-Za-z]{1})(_)([A-Z]{1}[A-Z0-9]{2})(_)?(0x[a-z0-9]{6})?',
    re.IGNORECASE | re.DOTALL)
TIMESTAMP_REGEX = re.compile("\\x15(\d+)_(\d+)\\x15")


def _region_boundaries_to_dataframe(region_lines):
    """
    Converts the region lines from a cha_structure file into a dataframe with three columns: region_type, start, end
    :param region_lines: a list of string read from the first part of a cha structure file
    :return: a pandas dataframe
    """
    boundaries_df = pd.DataFrame(columns=('region_type', 'which_boundary', 'time'),
                                 # Each line is "<region_type> <starts|ends> <timestamp>"
                                 data=[line.split() for line in region_lines])

    assert boundaries_df.region_type.isin(REGION_TYPES).all()

    # For each type, count all starts 1, 2, 3 and count all ends 1, 2, 3 so that we can then match starts to ends.
    boundaries_df['position'] = boundaries_df.groupby(['region_type', 'which_boundary']).cumcount() + 1
    # Match starts to ends and combine
    starts = boundaries_df[boundaries_df.which_boundary == 'starts'].drop(columns='which_boundary')
    ends = boundaries_df[boundaries_df.which_boundary == 'ends'].drop(columns='which_boundary')
    regions = pd.merge(left=starts, right=ends, how='left', on=['region_type', 'position'])
    assert starts.shape[0] == ends.shape[0] == regions.shape[0]

    # Rename and reorder columns
    regions = regions.rename(columns=dict(time_x='start', time_y='end'))[['region_type', 'start', 'end', 'position']]
    regions[['start', 'end']] = regions[['start', 'end']].astype(int)

    # Remove regions that have zero duration
    regions = regions[regions.start != regions.end]

    # Order by onset ascending, then offset descending, then region_type in alphabetical order
    regions = regions.sort_values(by=['start', 'end', 'region_type'],
                                  ascending=[True, False, True]).reset_index(drop=True)

    return regions


def _subregion_ranks_to_dataframe(subregion_rank_lines, subregion_count=DEFAULT_SUBREGION_COUNT):
    """
    Converts the subregion lines from a cha_structure file into a dataframe with two columns: position, rank
    :param subregion_rank_lines: a list of strings read from the second part of a cha structure file
    :param subregion_count: for known recordings with four subregions
    :return:
    """
    # Each row looks like "Position: 1, Rank: 1", so we can just extract position and rank using a regular expression
    subregion_ranks = (pd.Series(subregion_rank_lines)
                       .str.extractall(r'Position: (?P<position>\d+), Rank: (?P<subregion_rank>\d+)')
                       .reset_index(drop=True))

    # There should always be exactly five subregions and five ranks: 1 to 5
    positions = sorted(subregion_ranks.position.tolist())
    ranks = sorted(subregion_ranks.subregion_rank.tolist())
    assert positions == ranks == [str(i + 1) for i in range(subregion_count)]

    return subregion_ranks


def assert_numbers(*numbers):
    assert all((isinstance(number, int) or isinstance(number, float) for number in numbers))


def _set_difference_of_intervals(minuend, subtrahend):
    """
    Set-subtracts a closed interval from an open interval. The result is (a possibly empty) list of open intervals.
    :param minuend: (x1, x2) tuple of ints/floats representing the interval to be subtracted from
    :param subtrahend: (y1, y2) tuple of ints/floats representing the interval to be subtracted
    :return: list of 0, 1, or 2 paris of numbers in the natural order
    """
    x1, x2 = minuend
    y1, y2 = subtrahend
    assert_numbers(x1, x2, y1, y2)
    assert x1 < x2 and y1 < y2

    # Subtraction of (y1, y2) is equivalent to the union of subtracting (-Inf, y2] and [y1, Inf):
    # A \ (B1 ∧ B2) = (A \ B1) ∪ (A \ B2)
    # Further, if A = (x1, x2) and B1 = [y1, Inf), A \ B = (x1, min(x2, y1)) := (z1, z2) as long as z1 < z2
    result = [(z1, z2)
              for (z1, z2) in [(x1, min(x2, y1)),  # (x1, x2) \ [y1, Inf)
                               (max(x1, y2), x2)]   # (x1, x2) \ (-Inf, y2]
              if z1 < z2]

    return result


def _remove_interval_from_regions(regions, start, end):
    """
    Removes a time interval from each region in regions. As a results, each region can:
    - disappear totally if it contained within the removed interval,
    - get shortened on one side if only of it ends is within the remove interval,
    - become two shorter regions if the removed interval is fully contained within the region.
    :param regions: a regions dataframe (region_type, start, end columns)
    :param start: int/float, start of the interval to be removed
    :param end: int/float, end of the interval to be removed
    :return: a new dataframe with one row per each resulting region part or None, if all regions get removed
    """
    assert_numbers(start, end)
    with_interval_removed = regions.copy()

    # For each region, get a list of starts and ends of its subregions after the removal
    new_starts_and_ends = 'new_starts_and_ends'
    with_interval_removed[new_starts_and_ends] = with_interval_removed.apply(
        lambda row: _set_difference_of_intervals(minuend=(int(row.start), int(row.end)), subtrahend=(start, end)),
        axis='columns')

    # Now, each element in that list should get its own row.
    with_interval_removed = (with_interval_removed
                             .explode(new_starts_and_ends)
                             .dropna(subset=[new_starts_and_ends])  # Empty lists result in an NA row
                             .drop(columns=['start', 'end'])  # The original start and end can be dropped now
                             )

    # If no regions are left after the removal, return None
    if with_interval_removed.size == 0:
        return

    # Split 'new_start_and_ends' column that contains (start, end) tuples into two columns - start and end.
    with_interval_removed[['start', 'end']] = with_interval_removed.new_starts_and_ends.values.tolist()
    with_interval_removed.drop(columns=[new_starts_and_ends], inplace=True)

    # Finally, restore the original column order, reset index, and return
    return with_interval_removed[regions.columns].reset_index(drop=True)


def _remove_overlaps_from_other_regions(regions, dominant_region_type):
    """
    Takes regions of a single kind (e.g., silences) and removes from all other regions all overlapping parts (e.g., from
    each non-silence region removes any part that overlaps with any of the silence regions).
    :param regions: a dataframe with at least region_type, start, and end columns
    :param dominant_region_type: the region type overlaps with which will be removed from other regions.
    :return: copy of the regions dataframe with some of the non-dominant regions modified
    """
    is_dominant = regions.region_type == dominant_region_type

    # If there are no "other" regions, there is nothing to do. This only happens for recordings from months 6 and 7 that
    # only have skips or only have silences.
    if is_dominant.all():
        return regions

    dominant, nondominant = regions[is_dominant], regions[~is_dominant]
    for row in dominant.itertuples():
        nondominant = _remove_interval_from_regions(nondominant, int(row.start), int(row.end))
        if nondominant is None:
            break

    # Combine with the dominant regions and return
    return pd.concat([dominant, nondominant]).reset_index(drop=True)


def _remove_silences_and_skips(regions):
    """
    From each region removes any parts that overlap with silences, then with skips.
    See _remove_interval_from_regions for details of how the removing works.
    :param regions: a pandas dataframe output by _read_cha_structure, for example
    :return: a dataframe with skips and silence removed as regions and the corresponding interval removed from other
    regions
    """
    for region_type in (RegionType.SILENCE.value, RegionType.SKIP.value):
        regions = _remove_overlaps_from_other_regions(regions=regions, dominant_region_type=region_type)

    return regions


def _overlaps_with_interval(regions, start, end):
    return ~((regions.start >= end) | (regions.end <= start))


def _contains_nested(regions, start, end):
    return (regions.start <= start) & (end <= regions.end)


# TODO: This is unnecessarily complicated. The plan was to have multiple kinds of conditions: remove a subregion when it
#  strictly partially overlaps with one of the, say, surpluses, contain a surplus that is fully nested, is fully nested
#  within a surplus region, etc. The algorithm has been considerably simplified since then, so we only really need
#  _overlaps_with_interval and it is only ever used once on a single set of `other_region_types` (makeup/surplus). This
#  can all be simplified to something like `_remove_subregions_overlapping_with_makeup_or_surplus`.
#  If you get to this, it might be a good idea to break the process into logical, not technical steps: what we are
#  trying to do is figure out which subregions have been listened to: these are the ones with annotations or a special
#  comment except for those that only have annotations because they were used for makeup, which might have later been
#  renamed as surplus. So, it would make sense to have a `_remove_subregions_not_listened_to` function that would
#  account for annotations, special comments *and* overlapping makeup/surplus regions.
#  Another thing to consider: counting annotations after deoverlapping all the regions would be even better:
#  - annotations in makeup/surplus would not be in the subregions at all (overlaps are removed from subregions during
#    deoverlapping) so we would get the subregions without their own annotations removed for free,
#  - if those subregions have annotations outside of the makeup/surplus regions, we would probably want to know: in
#    such cases, the subregions would not get removed because they would still contain annotations ever after
#    deoverlapping.
#  Close issue https://github.com/BergelsonLab/blabpy/issues/12 when done.
def _remove_subregions(regions, condition_function, other_region_types):
    """
    Remove all subregions that satisfy a given condition depending on overlap with another region, e.g., have at least
    some overlap with silences or skips.
    :param regions: a full regions dataframe
    :param condition_function: a function that takes in regions, start, and end and returns a boolean Series that tells
    us whether each region in regions satisfies a given condition (e.g., partially overlaps, fully nested in, etc.)
    :param other_region_types: which regions should be tested against the condition? A list of RegionType properties.
    :return:
    """
    # Get necessary subsets of regions
    is_subregion = regions.region_type == RegionType.SUBREGION.value
    subregions, not_subregions = regions[is_subregion], regions[~is_subregion]
    # Convert other_region_types to a list of string to test against
    other_region_types_str = [other_region_type.value for other_region_type in other_region_types]
    other_regions = regions[regions.region_type.isin(other_region_types_str)]

    # Do the removal
    for other_region in other_regions.itertuples():
        condition_satisfied = condition_function(subregions, other_region.start, other_region.end)
        subregions = subregions[~condition_satisfied]

    # Combine with the other regions, restore order, return
    return pd.concat([subregions, not_subregions]).sort_index().reset_index(drop=True)


def _assign_makeup_and_extra_to_subregions(regions):
    return


def _aggregate_listen_time(regions, subregion_ranks):
    return


def _account_for_region_overlaps(regions):
    """
    Removes some subregions, modifies some other regions so that the resulting regions do not overlap and can be counted
    towards total listened time.
    1. Removes any subregions that overlap with any surplus region.
    2. Removes any subregions that contain in them nested makeup/surplus regions.
    3. Removes overlaps starting by removing overlaps with silences, then with skips, etc.
    :param regions: a regions dataframe such as the one output by _read_cha_structure
    :return:
    """
    # Some subregions need to be removed completely
    regions = _remove_subregions(regions, condition_function=_overlaps_with_interval,
                                 other_region_types=[RegionType.SURPLUS, RegionType.MAKEUP])

    # All other regions need to have parts of them remove dwhere they overlap with other regions.
    # The order matters, e.g. if you remove silences first, the silence will remain in their original form.
    # Deliberately manually placed regions should always trump the automatic ones.
    # Out of those, if a skip overlaps with any of the other ones, the skip should carry more weight and be left intact.
    # Makeup, surplus, extras should not overlap at all so their order does not matter.
    dominant_region_types = [RegionType.SKIP,
                             RegionType.SURPLUS, RegionType.MAKEUP, RegionType.EXTRA,
                             RegionType.SILENCE]
    # The list above should contain everything but subregions
    assert len(dominant_region_types) == len(REGION_TYPES) - 1
    # Check that surplus, makeup, and extra regions do not overlap
    _assert_no_overlaps(regions[regions.region_type.isin(
        [RegionType.SURPLUS.value, RegionType.MAKEUP.value, RegionType.EXTRA.value])])

    for dominant_region_type in dominant_region_types:
        regions = _remove_overlaps_from_other_regions(regions, dominant_region_type.value)
    return regions


def _remove_subregions_without_annotations(regions_df, listened_but_empty):
    """
    This function has to be run before any region adjustments because it does not account for possible splits resulting
    from having, for example, a skip in the middle.
    :param regions_df: a regions dataframe
    :param listened_but_empty: list of additional timestamp offset corresponding to the special comments that mark
    subregions that were listened to but did not have any codeable words
    :return: regions with possibly a few rows removed
    """
    # Check that no subregions have been split yet. Split subregions would result in duplicate position values.
    assert regions_df[regions_df.region_type == RegionType.SUBREGION.value].duplicated(subset=['position']).sum() == 0

    # Which regions contain a "listened but empty" comment?
    has_listened_but_empty_comment = regions_df.apply(
        lambda row: any(row.start <= timestamp <= row.end for timestamp in listened_but_empty),
        axis='columns')

    # Remove subregions without annotations unless they have a special comment
    is_subregion = regions_df.region_type == RegionType.SUBREGION.value
    has_annotations = regions_df.annotation_count > 0
    should_be_removed = is_subregion & ~(has_annotations | has_listened_but_empty_comment)
    regions_df = regions_df[~should_be_removed]

    return regions_df.drop(columns='annotation_count')


def _assert_no_overlaps(regions):
    """
    Assert that none of the regions overlap, being back-to-back is fine.
    :param regions: a pandas DataFrame with start and end columns
    :return:
    """
    boundaries = regions.sort_values(by=['start', 'end'], ascending=[True, False])[['start', 'end']]
    previous_end = boundaries.end.shift(fill_value=-1)
    assert (boundaries.start >= previous_end).all()


def _total_eligible_time(regions):
    """
    Sums up duration of all regions except for silences, skips and surpluses. Assumes that the regions have been
    de-overlapped.
    :param regions: a regions dataframe that has already been de-overlapped
    :return: total duration as an integer
    """
    _assert_no_overlaps(regions)
    region_types_to_exclude = [RegionType.SURPLUS.value, RegionType.SILENCE.value, RegionType.SKIP.value]
    total_time_per_region = _total_time_and_count_per_region_type(regions_df=regions[
        ~regions.region_type.isin(region_types_to_exclude)])
    return total_time_per_region.total_time.sum()


def _process_regions(regions, annotation_timestamps, listened_but_empty):
    """
    Processes regions:
    - removes subregions without annotations (unless they contain a special comment instead),
    - removes any overlaps between different regions in a certain order,
    :param regions:
    :param annotation_timestamps:
    :param listened_but_empty:
    :return:
    """
    # Do not count subregions without annotations, unless they contain a special comment
    regions = _remove_subregions_without_annotations(regions, annotation_timestamps, listened_but_empty)

    # Account for region overlaps
    regions = _account_for_region_overlaps(regions)

    return regions


def _total_time_and_count_per_region_type(regions_df):
    """
    Calculates total duration and region count for each region type. Counts split regions only once.
    :param regions_df: a regions dataframe with columns region_type, start, end, and position
    :return:
    """
    return (regions_df
            .assign(duration=(lambda df: df.end - df.start))
            .groupby('region_type')
            .aggregate(total_time=('duration', 'sum'),
                       region_count=('position', 'nunique')))


def _extract_timestamps(clan_file_text: str):
    """
    Extract all timestamps from a clan file remembering their positions in text. Raises a ValueError if the onsets or
    offsets are not monotonic increasing.
    :param clan_file_text: clan file as one long string
    :return: a dataframe with one row per timestamp and three columns: onset, offset, 'position_in_text'
    """
    timestamps = pd.DataFrame(
        columns=['onset', 'offset', 'position_in_text'],
        data=[(int(match.group(1)), int(match.group(2)), match.start())
              for match in TIMESTAMP_REGEX.finditer(clan_file_text)],
        dtype=int)

    if not timestamps.onset.is_monotonic:
        # TODO: uncomment or delete once timestamp inconsistencies have been dealt with
        # not (timestamps.onset.is_monotonic and timestamps.offset.is_monotonic and
        # (timestamps.onset <= timestamps.offset).all()):
        raise ValueError('Timestamps are not in the right order.')

    return timestamps


def _match_with_timestamps(positions_in_text, timestamps_df, above=None, below=None):
    """
    For each row in a DataFrame/Series that has a 'position_in_text' integer column finds the last timestamp before that
    position (if before is True) or the first one after (if after is True).
    :param positions_in_text: an arbitrary dataframe with a 'position_in_text' integer column which must be monotonic
    increasing
    :param timestamps_df: same, but normally it is a dataframe as output by _extract_timestamps
    :param above: if True, look for the last timestamp above
    :param below: if True, look for the first timestamp below
    :return: a copy of df_with_position_in_text with two additional columns: onset and offset
    """
    if below and not above:
        direction = 'forward'
    elif above and not below:
        direction = 'backward'
    else:
        raise ValueError('Exactly one of `above` and `below` arguments must evaluate to True')

    # Check that the positions are monotonic increasing
    if isinstance(positions_in_text, pd.Series):
        assert positions_in_text.name == 'position_in_text'
        assert positions_in_text.is_monotonic
    elif isinstance(positions_in_text, pd.DataFrame):
        assert positions_in_text.position_in_text.is_monotonic
    assert timestamps_df.position_in_text.is_monotonic

    return pd.merge_asof(positions_in_text, timestamps_df,
                         on='position_in_text', direction=direction)


def _extract_annotation_timestamps(clan_file_text: str):
    """
    Find all annotation timestamps in a clan/cha file
    :param clan_file_text: string with the clan file text
    :return: a pandas dataframe with two columns: 'onset' and 'offset'; and one row per each annotation found
    """
    annotation_positions_in_text = pd.Series([match.start() for match in ANNOTATION_REGEX.finditer(clan_file_text)],
                                             name='position_in_text')

    # Add the first timestamps below the annotations in the file
    annotation_timestamps = _match_with_timestamps(
        positions_in_text=annotation_positions_in_text,
        timestamps_df=_extract_timestamps(clan_file_text),
        below=True)

    # Here, we are only interested in unique timestamps, not unique annotations, so we should remove the duplicates
    annotation_timestamps = (annotation_timestamps
                             [['onset', 'offset']]
                             .drop_duplicates(keep='first')
                             .reset_index(drop=True))

    return annotation_timestamps


def _add_per_region_timestamp_count(regions_df, timestamps):
    """
    Count annotations that start and end within each subregions
    :param regions_df: a dataframe with 'start' and 'end' numeric columns
    :param timestamps: a dataframe with 'onset' column
    :return: regions_df with an additional column 'annotation_count'
    """
    regions_df_columns = regions_df.columns.tolist()
    # Brute-force solution: take a cross-product of regions and annotations and filter out rows where annotation is not
    # within region boundaries
    with_annotation_counts = (
        regions_df
        # There is no cross join in pandas AFAIK, so we'll have to join on a dummy constant column
        .assign(cross_join=0)
        .merge(timestamps.assign(cross_join=0), on='cross_join')
        # The onset should within region boundaries
        .query('start <= onset and onset < end')
        .groupby(regions_df_columns)
        .size()
        .rename('annotation_count')
        .reset_index()
        # Above, we lost regions that do not have any annotations in them, let's put them back with the count of 0
        .merge(regions_df, on=regions_df_columns, how='right')
        .fillna(dict(annotation_count=0)))

    return with_annotation_counts


def milliseconds_to_hours(ms):
    return round(ms / (60 * 60 * 1000), PRECISION)


def _extract_subregion_info(comment):
    """
    Extracts subregion position, rank, offset from a comment line from a clan (cha) file.
    :param comment: a string from the subregion boundary comment in a cha file
    :return: position, rank, offset
    """
    subregion_position_regex = re.compile(r'subregion (\d+) of (\d+)')
    subregion_rank_regex = re.compile(r'ranked (\d+) of (\d+)')
    subregion_time_regex = re.compile(r'at (\d+)')
    position = subregion_position_regex.search(comment).group(1)
    rank = subregion_rank_regex.search(comment).group(1)
    offset = int(subregion_time_regex.findall(comment)[0])

    return position, rank, offset


def _extract_region_info(clan_file_text: str, subregion_count=DEFAULT_SUBREGION_COUNT):
    """
    Extracts region boundaries, subregions ranks, and info about "listened to, nothing to annotate" from a clan file.
    :param clan_file_text: string with the clan file text
    :return:
    """
    # Find all the comments except for the LENA comments

    # This regex will consume any number of lines the first of which starts with '%com' or '%xcom' and the last of which
    # is followed by a line that starts with any character except for the tabulation symbol. This is necessary to
    # consume the multiline comments which use `\t` as the line continuation symbol. For this regex to work, we need
    # to additionally specify the MULTILINE (we want '^' to match the beginning of any line, not just of the whole text)
    # and the DOTALL (we want '.*' to be able to cross the line boundaries) flags.
    comment_line_regex = r'^%x?com:.*?(?=^[^\t])'
    comments_df = pd.DataFrame(
        columns=('text', 'position_in_text'),
        data=[(match.group(), match.start())
              for match in re.finditer(comment_line_regex, clan_file_text, flags=re.MULTILINE + re.DOTALL)])
    # Remove LENA comments, counting '|' solution comes from pyclan
    comments_df = comments_df[comments_df.text.str.count('\|') <= 3]
    # Add timestamp info
    comments_df = _match_with_timestamps(
        positions_in_text=comments_df,
        timestamps_df=_extract_timestamps(clan_file_text),
        above=True)
    # Some files have region comments before the first tier and thus they don't get a timestamp, so it will have
    # NaN as offset, which will force pandas to convert the whole column to 'float64' which will cause problems down the
    # line. So, here, we will convert offset to a special integer datatype that supports NaNs and will fill the ones in
    # the beginning with 0.
    for column in ('onset', 'offset'):
        comments_df[column] = comments_df[column].astype(pd.Int64Dtype())
        # .ffill() will ensure we are only filling NaNs in the beginning.
        comments_df.loc[comments_df[column].ffill().isnull(), column] = 0

    subregions = []  # List of strings of the format 'Position: {}, Rank: {}'
    region_boundaries = []  # List of strings of the format
    listened_but_empty = []  # List of integers - offsets of the corresponding comments

    # Code below is copied from annot_distr
    for comment_row in comments_df.itertuples():
        comment = comment_row.text
        row_offset = comment_row.offset
        if 'subregion' in comment:
            if NO_CODEABLE_WORDS_BUT_LISTENED_COMMENT in comment:
                listened_but_empty.append(row_offset)
            else:
                sub_pos, sub_rank, offset = _extract_subregion_info(comment)
                if 'starts' in comment:
                    region_boundaries.append(('subregion starts', offset))
                # Only adding after ends in order to not add the position and rank info twice to the subregions list.
                elif 'ends' in comment:
                    region_boundaries.append(('subregion ends', offset))
                    subregions.append('Position: {}, Rank: {}'.format(sub_pos, sub_rank))

            continue

        if 'extra' in comment:
            if 'begin' in comment:
                region_boundaries.append(('extra starts', row_offset))
            elif 'end' in comment:
                region_boundaries.append(('extra ends', row_offset))
        elif 'silence' in comment:
            if 'start' in comment:
                region_boundaries.append(('silence starts', row_offset))
            elif 'end' in comment:
                region_boundaries.append(('silence ends', row_offset))
        elif 'skip' in comment:
            if 'begin' in comment:
                region_boundaries.append(('skip starts', row_offset))
            elif 'end' in comment:
                region_boundaries.append(('skip ends', row_offset))
        elif 'makeup' in comment or 'make-up' in comment or 'make up' in comment:
            if 'begin' in comment:
                region_boundaries.append(('makeup starts', row_offset))
            elif 'end' in comment:
                region_boundaries.append(('makeup ends', row_offset))
        elif 'surplus' in comment:
            if 'begin' in comment:
                region_boundaries.append(('surplus starts', row_offset))
            elif 'end' in comment:
                region_boundaries.append(('surplus ends', row_offset))

    # The code below emulates reading from cha_structures files that annot_distr creates
    region_boundaries_df = _region_boundaries_to_dataframe([' '.join(map(str, rb)) for rb in region_boundaries])
    if subregion_count > 0:
        subregion_ranks_df = _subregion_ranks_to_dataframe(subregions, subregion_count=subregion_count)
    else:
        subregion_ranks_df = None

    return region_boundaries_df, subregion_ranks_df, listened_but_empty


def _preprocess_region_info(clan_file_text: str, subregion_count=DEFAULT_SUBREGION_COUNT):
    """
    Extract and preprocess region info from the cha files.
    :param clan_file_text: contents of the clan files as a string
    :param subregion_count: expected total number of subregions. Should be 0 for months 6 and 7, and 5 for all other
     months except for a few known exceptions.
    :return: (regions_raw, regions_processed, subregion_ranks_df, listened_but_empty) where:
        - regions_raw - dataframe with all the regions (possibly overlapping), their onsets, offsets, position within
            regions of the same kind, and the number of annotations in each of them (same annotation can count towards
            multiple regions if they are overlapping),
        - regions_processed - regions_raw with some regions removed and then deoverlapped,
        - subregion_ranks_df - dataframe with positions and ranks of subregions,
        - listened_but_empty - list of comments used to mark subregions that were listened to but didn't have any
            codable objects and thus don't have any annotations.
    """
    annotation_timestamps = _extract_annotation_timestamps(clan_file_text)
    regions_raw, subregion_ranks_df, listened_but_empty = _extract_region_info(
        clan_file_text, subregion_count=subregion_count)

    # Process regions. For months 6 and 7 that have no subregions, we only need to remove the overlapping parts of
    # the regions.
    regions_raw = _add_per_region_timestamp_count(regions_df=regions_raw,
                                                  timestamps=annotation_timestamps)
    if subregion_count > 0:
        regions_processed = _remove_subregions_without_annotations(regions_raw, listened_but_empty)
    else:
        regions_processed = regions_raw.copy()
    regions_processed = _account_for_region_overlaps(regions_processed)

    return regions_raw, regions_processed, subregion_ranks_df, listened_but_empty


def listen_time_stats_for_report(clan_file_text: str, subregion_count=DEFAULT_SUBREGION_COUNT):
    """
    Caculates a number of listen time statistics for the rmd report in the annot_distr repository.
    :param clan_file_text: string with the clan file text
    :param subregion_count: the number of subregions to expect, most have 5 but some have 4. Months 6 and 7 have zero.
    :return:
    """
    # Extract the necessary information from the clan file
    regions_raw, regions_processed, subregion_ranks_df, _ = _preprocess_region_info(clan_file_text=clan_file_text,
                                                                                    subregion_count=subregion_count)

    # Calculate total for each region type
    totals_raw = _total_time_and_count_per_region_type(regions_raw)
    totals_processed = _total_time_and_count_per_region_type(regions_processed)

    # Total time and count for regions that have been added manually. In case there is a makeup region from 0 to 10
    # and there is a skip inside of it from 2 to 8, we want to count it as 1 region with the listened time of 2 + 2 = 4.
    # Therefore, the adding up is based on processed regions, _total_time_and_count_per_region_type already accounts for
    # the fact that this makeup region will have been split into two (0-2 and 8-10).
    additional_regions = (RegionType.MAKEUP.value, RegionType.EXTRA.value, RegionType.SURPLUS.value)
    stats = {f'num_{region_type}_region': totals_processed.region_count.get(region_type, 0)
             for region_type in additional_regions}
    stats.update({f'{region_type}_time': totals_processed.total_time.get(region_type, 0)
                 for region_type in additional_regions})

    # Same for the subregions. The count field is named differently for historical reasons.
    stats['subregion_time'] = totals_processed.total_time.get(RegionType.SUBREGION.value, 0)
    stats['num_subregion_with_annot'] = totals_processed.region_count.get(RegionType.SUBREGION.value, 0)

    # These stats are relevant for total time calculation for months 6 and 7 which were listened to in full. The
    # overlap is actually no longer relevant because overlaps are now removed during processing.
    stats.update(dict(skip_silence_overlap_hour=0,
                      skip_time=totals_processed.total_time.get(RegionType.SKIP.value, 0),
                      silence_time=totals_processed.total_time.get(RegionType.SILENCE.value, 0),
                      silence_raw_hour=milliseconds_to_hours(totals_raw.total_time.get(RegionType.SILENCE.value, 0))))
    # Note: this can probably be optimized because we already extracted timestamps once - when we looked for annotation
    # timestamps
    last_timestamp_offset = _extract_timestamps(clan_file_text).iloc[-1].offset
    stats['end_time'] = last_timestamp_offset

    # The total listen time calculation depends on whether the full recording was listened to (month 6 and 7, excluding
    # silences and skips) or just the subregions (months 8+, additionally makeup, extra, and surplus regions
    if subregion_count > 0:
        stats['total_listen_time'] = _total_eligible_time(regions=regions_processed)
    else:
        stats['total_listen_time'] = stats['end_time'] - stats['skip_time'] - stats['silence_time']

    # Subregion positions and ranks
    if subregion_count > 0:
        stats['positions'] = subregion_ranks_df.position.to_list()
        stats['ranks'] = subregion_ranks_df.subregion_rank.to_list()
    else:
        stats['positions'], stats['ranks'] = list(), list()

    # Stats before processing
    stats['subregion_raw_hour'] = milliseconds_to_hours(totals_raw.total_time.get(RegionType.SUBREGION.value, 0))
    stats['num_raw_subregion'] = milliseconds_to_hours(totals_raw.total_time.get(RegionType.SUBREGION.value, 0))

    # This is not exactly correct: annotations that share their timestamp are only counted once.
    annotation_counts_raw = (regions_raw
                             [regions_raw.region_type == RegionType.SUBREGION.value]
                             .sort_values(by='position')
                             .annotation_count
                             .astype(int)
                             .to_list())
    stats['annotation_counts_raw'] = annotation_counts_raw

    return stats


def _get_subregion_count(child, month):
    if (child, month) in RECORDINGS_WITH_FOUR_SUBREGIONS:
        return 4
    elif month in (6, 7):
        return 0
    else:
        # It is 5, of course, but it is used so often in this module that hard-coding it was not an option.
        return DEFAULT_SUBREGION_COUNT



import argparse
from pathlib import Path

from blabpy.vihi.intervals.intervals import batch_create_files_with_random_regions


def cli_batch_create_files_with_random_regions(args=None):
    """
    Run batch_create_files_with_random_regions from command line. See cli arguments below.
    :param args: for testing only
    :return:
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('info_spreadsheet_path', help='Path to the info_spreadsheet.csv file.')
    parser.add_argument('seed', nargs='?', default=None, help='Optional seed argument. Used mostly for testing.')
    args = parser.parse_args(args)
    batch_create_files_with_random_regions(Path(args.info_spreadsheet_path), seed=args.seed)


if __name__ == "__main__":
    cli_batch_create_files_with_random_regions()



from ..paths import get_pn_opus_path


POPULATIONS = ['VI', 'HI', 'TD']


def get_vihi_path():
    """
    Finds the path to the VIHI folder on PN-OPUS
    :return: Path object
    """
    return get_pn_opus_path() / 'VIHI'


def get_subject_files_path():
    """
    Returns the path to the SubjectFiles folder within the VIHI folder
    :return: Path object
    """
    return get_vihi_path() / 'SubjectFiles'


def _get_modality_path(modality):
    """
    Return the path to a modality folder withing the SubjectFiles folder
    :return: Path object
    """
    return get_subject_files_path() / modality


def get_lena_path():
    """
    Returns the path to the LENA folder within the SubjectFiles folder
    :return: Path object
    """
    return _get_modality_path('LENA')


def _id_from_int(id_):
    """
    Converts integer subject and recordings ids to a 3-digit-long zero-paddes string
    :param id_: int
    :return: str
    """
    return f'{id_:03}'


def _check_id_string(id_):
    """
    Checks that the subject or recordings id is formatted correctly by converting to number and back.
    :param id_:
    :return:
    """
    assert isinstance(id_, str)
    assert _id_from_int(int(id_)) == id_


def _check_population(population):
    assert population in POPULATIONS


def compose_full_recording_id(population: str, subject_id: str, recording_id: str):
    """
    Combines population type, subject id, and recording id into a full recording id, e.g., VI_123_456
    :param population: VI/HI/TD
    :param subject_id:
    :param recording_id:
    :return:
    """
    _check_population(population)
    _check_id_string(subject_id)
    _check_id_string(recording_id)
    return f'{population}_{subject_id}_{recording_id}'


def parse_full_recording_id(full_recording_id):
    """
    Parse a full recording id, e.g., VI_123_456, into three constituent parts.
    :param full_recording_id: a string
    :return: a dict with string values and keys population, subject_id, recording_id
    """
    population, subject_id, recording_id = full_recording_id.split('_')
    _check_population(population)
    _check_id_string(subject_id)
    _check_id_string(recording_id)

    return dict(population=population, subject_id=subject_id, recording_id=recording_id)


def get_lena_population_path(population):
    """
    Find the LENA population-level folder
    :param population: one of POPULATIONS
    :return: Path object
    """
    _check_population(population)
    return get_lena_path() / population


def get_lena_subject_path(population, subject_id):
    """
    Find the LENA subject-level folder
    :param population: one of POPULATIONS
    :param subject_id: zero-padded three-digit string
    :return: Path object
    """
    _check_population(population)
    _check_id_string(subject_id)
    return get_lena_population_path(population) / f'{population}_{subject_id}'


def get_lena_recording_path(population, subject_id, recording_id):
    """
    Find the LENA subject-level folder
    :param population: one of POPULATIONS
    :param subject_id: zero-padded three-digit string
    :param recording_id: --//--
    :return: Path object
    """
    _check_population(population)
    _check_id_string(subject_id)
    _check_id_string(recording_id)
    return get_lena_subject_path(population, subject_id) / compose_full_recording_id(population, subject_id, recording_id)


def get_its_path(population, subject_id, recording_id):
    """
    Find the .its file.
    :param population: one of POPULATIONS
    :param subject_id: zero-padded three-digit string
    :param recording_id: --//--
    :return: Path object
    """
    _check_population(population)
    _check_id_string(subject_id)
    _check_id_string(recording_id)

    subject_dir = get_lena_subject_path(population, subject_id)
    assert subject_dir.exists(), 'Can\'t find the subject folder. Check the arguments.'

    full_recording_id = compose_full_recording_id(population, subject_id, recording_id)
    its_path = subject_dir / full_recording_id / f'{full_recording_id}.its'

    return its_path


def get_rttm_path(population, subject_id, recording_id):
    """
    Find the .rttm file with the VTC output for a single recording.
    :param population: one of POPULATIONS
    :param subject_id: zero-padded three-digit string
    :param recording_id: --//--
    :return: Path object
    """
    _check_population(population)
    _check_id_string(subject_id)
    _check_id_string(recording_id)

    lena_path = get_lena_path()
    full_recording_id = compose_full_recording_id(population, subject_id, recording_id)
    rttm_path = lena_path / 'derivatives' / 'vtc' / 'separated' / full_recording_id / 'all.rttm'

    return rttm_path


def get_eaf_path(population, subject_id, recording_id):
    """
    Find the .eaf file with the ACLEW-style annotations for a single recording.
    :param population: one of POPULATIONS
    :param subject_id: zero-padded three-digit string
    :param recording_id: --//--
    :return: Path object
    """
    _check_population(population)
    _check_id_string(subject_id)
    _check_id_string(recording_id)

    subject_dir = get_lena_subject_path(population, subject_id)
    assert subject_dir.exists(), 'Can\'t find the subject folder. Check the arguments.'

    full_recording_id = compose_full_recording_id(population, subject_id, recording_id)
    eaf_path = subject_dir / full_recording_id / f'{full_recording_id}.eaf'

    return eaf_path



import random
import shutil
from pathlib import Path

import numpy as np
import pandas as pd

from blabpy.eaf import EafPlus
from blabpy.utils import OutputExistsError
from blabpy.vihi.intervals import templates
from blabpy.vihi.paths import get_lena_recording_path, parse_full_recording_id
from blabpy.utils import df_to_list_of_tuples

_ms_in_a_minute = 60 * 10**3
CONTEXT_BEFORE = 2 * _ms_in_a_minute
CODE_REGION = 2 * _ms_in_a_minute
CONTEXT_AFTER = 1 * _ms_in_a_minute
INTERVALS_FOR_ANNOTATION_COUNT = 15
INTERVALS_EXTRA_COUNT = 5


def _overlap(onset1, onset2, width):
    """
    Do the width-long intervals starting with onset1 and onset2 overlap?
    (1, 2) and (2, 3) are considered to be non-overlapping.
    :param onset1: int, onset1 > 0, start of the first interval,
    :param onset2: int, onset2 != onset1 & onset2 > 0, start of the second interval,
    :param width: int, width > 0, their common duration
    :return: True/False
    """
    if onset2 < onset1 < onset2 + width:
        return True
    elif onset2 - width < onset1 < onset2:
        return True
    return False


def select_intervals_randomly(total_duration, n=5, t=5, start=30, end=10):
    """
    Randomly selects n non-overlapping regions of length t that start not earlier than at minute start and not later
    than (total_duration - end).
    int total_duration: length of recording in minutes
    int n: number of random intervals to choose
    int t: length of region of interest (including context)
    int start: minute at which the earliest interval can start
    return: a list of (onset, offset + t) tuples
    """
    candidate_onsets = list(range(start, min(total_duration - t, total_duration - end)))
    random.shuffle(candidate_onsets)
    selected_onsets = []
    for possible_onset in candidate_onsets:
        # Select onsets until we have the required number of intervals
        if len(selected_onsets) >= n:
            break
        # Check that the candidate region would not overlap with any of the already selected ones
        if not any(_overlap(possible_onset, selected_onset, t) for selected_onset in selected_onsets):
            selected_onsets.append(possible_onset)

    return [(onset, onset + t) for onset in selected_onsets]


def two_lists_of_intervals_overlap(list1, list2):
    """
    Checks whether two lists of intervals (order tuples of numbers)
    :param list1: list 1
    :param list2: list 2
    :return: boolean
    """
    def overlaps_with_list2(start1, end1):
        return any(start1 < end2 and end1 > start2
                   for start2, end2
                   in list2)

    return any(overlaps_with_list2(start1, end1)
               for start1, end1
               in list1)


# TODO: this function should take a list/dataframe with both code and context region boundaries, the current way is kind
#  of backwards.
def add_annotation_intervals_to_eaf(eaf, intervals):
    """
    Adds annotation intervals to an EafPlus object. The input is list of *full* intervals - including the context.
    :param eaf: EafPlus objects with tiers added, assumed to be empty
    :param intervals: dataframe with at least the following columns:
        'code_onset_wav', 'code_offset_wav', 'context_onset_wav', 'context_offset_wav', 'sampling_type'
    :return: EafPlus object with intervals added.
    """
    # Check for overlaps with existing code intervals
    existing_code_intervals_list = eaf.get_time_intervals('code')
    new_code_intervals_list = df_to_list_of_tuples(intervals[['code_onset_wav', 'code_offset_wav']])
    assert not two_lists_of_intervals_overlap(existing_code_intervals_list, new_code_intervals_list)

    # Figure out which code_num we should start with (it is last_code_num + 1)
    try:
        code_num_values = eaf.get_values('code_num')
        existing_code_nums = [int(code_num) for code_num in code_num_values]
    except ValueError as e:
        msg = f'All code num values should be integers, some of the following weren\'t\n{code_num_values}'
        raise ValueError(msg) from e

    last_code_num = 0 if len(existing_code_nums) == 0 else max(existing_code_nums)
    n_intervals_to_add = intervals.shape[0]
    new_code_nums = list(range(last_code_num + 1, last_code_num + n_intervals_to_add + 1))

    # Sort new intervals by onset and assign code_num
    intervals = (intervals
                 .sort_values(by=['code_onset_wav', 'code_offset_wav'])
                 .assign(code_num=new_code_nums,
                         on_off=lambda df: df.code_onset_wav.astype(str) + '_' + df.code_offset_wav.astype(str)))

    # Add intervals
    for _, row in intervals.iterrows():
        eaf.add_annotation("code", row.code_onset_wav, row.code_offset_wav)
        eaf.add_annotation("code_num", row.code_onset_wav, row.code_offset_wav, value=str(row.code_num))
        eaf.add_annotation("on_off", row.code_onset_wav, row.code_offset_wav, value=row.on_off)
        eaf.add_annotation("context", row.context_onset_wav, row.context_offset_wav)
        eaf.add_annotation("sampling_type", row.code_onset_wav, row.code_offset_wav, value=row.sampling_type)

    return eaf, intervals


def prepare_eaf_from_template(etf_template_path):
    """
    Loads eaf template, adds empty tiers and returns an EafPlus object ready for inserting annotation interval data.
    :param etf_template_path:
    :return: EafPlus object
    """
    # Load
    eaf = EafPlus(etf_template_path)

    # Add tiers
    transcription_type = "transcription"
    tier_ids = ('code', 'context', 'code_num', 'on_off', 'sampling_type')
    for tier_id in tier_ids:
        eaf.add_tier(tier_id=tier_id, ling=transcription_type)

    return eaf


def create_eaf_from_template(etf_template_path, intervals):
    """
    Writes an eaf file <id>.eaf to the output_dir by adding intervals to the etf template at etf_path.
    :param etf_template_path: path to the .etf template file
    :param intervals: dataframe with at least the following columns:
        'code_onset_wav', 'code_offset_wav', 'context_onset_wav', 'context_offset_wav'
    :return: an EafPlus objects with the code, code_num, on_off, and context annotations.
    code_num is the number of interval within the interval_list
    context onset and offset are those from the intervals_list - it includes the region to annotate
    """
    eaf = prepare_eaf_from_template(etf_template_path)
    eaf, intervals = add_annotation_intervals_to_eaf(eaf=eaf, intervals=intervals)
    return eaf, intervals


def _region_output_files(full_recording_id):
    """
    Find the recording folder and list the output files as a dict.
    Factored out so we can check which files are already present during batch processing without creating random regions
    for the recordings that haven't been processed yet.
    :param full_recording_id:
    :return:
    """
    output_dir = get_lena_recording_path(**parse_full_recording_id(full_recording_id))
    output_filenames = {
        'eaf': f'{full_recording_id}.eaf',
        'pfsx': f'{full_recording_id}.pfsx',
        'csv': f'selected_regions.csv'
    }
    return {extension: Path(output_dir) / filename
            for extension, filename in output_filenames.items()}


def create_files_with_random_regions(full_recording_id, age, length_of_recording):
    """
    Randomly samples INTERVALS_FOR_ANNOTATION_COUNT five-min long regions to be annotated and creates three files:
    - <full_recording_id>.eaf - ELAN file with annotations prepared for the sampled intervals,
    - <full_recording_id>.pfsx - ELAN preferences file,
    - <full_recording_id>_selected-regions.csv - a table with onset and offsets of the selected regions.
    Raises an OutputExistsError if any of the files already exist.
    :param full_recording_id: full recording id, e.g. 'TD_123_456'
    :param age: age in months - will be used to select an .etf template
    :param length_of_recording: length of the actual file in minutes
    :return: None, writes files to the recording folder in VIHI
    """
    # check that none of the output files already exist
    output_file_paths = _region_output_files(full_recording_id=full_recording_id)
    paths_exist = [path for path in output_file_paths.values() if path.exists()]
    if any(paths_exist):
        raise OutputExistsError(paths=paths_exist)

    # select random intervals
    timestamps = select_intervals_randomly(int(length_of_recording), n=INTERVALS_FOR_ANNOTATION_COUNT)
    timestamps = [(x * 60000, y * 60000) for x, y in timestamps]
    timestamps.sort(key=lambda tup: tup[0])
    context_onsets, context_offsets = zip(*timestamps)
    intervals = pd.DataFrame.from_dict(dict(context_onset_wav=context_onsets, context_offset_wav=context_offsets))
    intervals.insert(0, 'code_onset_wav', intervals.context_onset_wav + CONTEXT_BEFORE)
    intervals.insert(1, 'code_offset_wav', intervals.context_offset_wav - CONTEXT_AFTER)
    intervals.insert(0, 'full_recording_id', full_recording_id)
    intervals.insert(1, 'sampling_type', 'random')

    # retrieve correct templates for the age
    etf_template_path, pfsx_template_path = templates.choose_template(age_in_months=age)

    # create an eaf object with the selected regions
    eaf, intervals = create_eaf_from_template(etf_template_path, intervals)

    # create the output files
    # eaf with intervals added
    eaf.to_file(output_file_paths['eaf'])
    # copy the pfsx template
    shutil.copy(pfsx_template_path, output_file_paths['pfsx'])
    # csv with the list of selected regions
    intervals.to_csv(output_file_paths['csv'], index=False)


def batch_create_files_with_random_regions(info_spreadsheet_path, seed=None):
    """
    Reads a list of recordings for which eafs with randomly selected regions need to be created. Outputs an eaf, a pfsx,
    and a *_selected_regions.csv files for each recording.
    If any of the output files for any of the recordings already exist, the process is aborted.

    :param info_spreadsheet_path: path to a csv that has the following columns:
     `age` with the child's age in months at the time of the recording,
     `length_of_recording` in minutes,
     `id`: recording identifier, such as VI_018_924
    :param seed: int, optional, random seed to be set before selecting random regions. Set only once, before processing
     all the recordings. For testing purposes mostly.
    :return: None
    """
    if seed:
        random.seed(seed)

    recordings_df = pd.read_csv(info_spreadsheet_path)

    # Check that the output files don't yet exist
    def some_outputs_exist(full_recording_id_):
        return any(path.exists() for path in _region_output_files(full_recording_id=full_recording_id_).values())
    recordings_previously_processed = recordings_df.id[recordings_df.id.apply(some_outputs_exist)]
    if recordings_previously_processed.any():
        msg = ('The following recordings already have random region files:\n'
               + '\n'.join(recordings_previously_processed)
               + '\nAborting!')
        raise FileExistsError(msg)

    # Create random regions
    for _, recording in recordings_df.iterrows():
        create_files_with_random_regions(full_recording_id=recording.id, age=recording.age,
                                         length_of_recording=recording.length_of_recording)
        print(f'{recording.id}: random regions created.')


def calculate_energy_in_one_interval(start, end, audio, low_freq: int = 0, high_freq: int = 100000):
    """
    Calculates energy from start to end from a recording loaded into memory.
    NB: The code is copied almost verbatim from ChildProject's energy-based sampler code.
    :param high_freq: upper frequency
    :param low_freq: lower frequency limit
    :param start: start in milliseconds
    :param end: end in milliseconds
    :param audio: pydub.AudioSegment object
    :return: float - energy in the interval
    """
    sampling_frequency = int(audio.frame_rate)

    def compute_energy_loudness(single_channel_chunk):
        if low_freq > 0 or high_freq < 100000:
            chunk_fft = np.fft.fft(single_channel_chunk)
            freq = np.abs(np.fft.fftfreq(len(chunk_fft), 1 / sampling_frequency))
            chunk_fft = chunk_fft[(freq > low_freq) & (freq < high_freq)]
            return np.sum(np.abs(chunk_fft) ** 2) / len(single_channel_chunk)
        else:
            return np.sum(single_channel_chunk ** 2)

    channels = audio.channels
    max_value = 256 ** (int(audio.sample_width)) / 2 - 1

    chunk = audio[start:end].get_array_of_samples()
    channel_energies = np.zeros(channels)

    for channel in range(channels):
        channel_chunk = np.array(chunk[channel::channels]) / max_value
        channel_energies[channel] = compute_energy_loudness(single_channel_chunk=channel_chunk)

    energy = np.sum(channel_energies)
    return energy


def calculate_energy_in_all_intervals(intervals, audio, low_freq: int = 0, high_freq: int = 100000):
    """
    Calculates energy in audio for each interval in intervals.
    :param high_freq: see calculate_energy_in_one_interval
    :param low_freq: see calculate_energy_in_one_interval
    :param intervals: a pandas dataframe containing "start" and "end" columns in seconds
    :param audio: pydub.AudioSegment object
    :return: a pandas Series object
    """
    return intervals.apply(lambda row:
                           calculate_energy_in_one_interval(start=row.start, end=row.end, audio=audio,
                                                            low_freq=low_freq, high_freq=high_freq),
                           axis='columns')


def _make_intervals_for_sub_recording(first_code_onset, last_code_offset, first_code_onset_wav):
    """
    Creates a sequence of intervals for one sub-recording.
    :param first_code_onset: datetime, onset of the first code region
    :param last_code_offset: datetime, offset of the first code region
    :param first_code_onset_wav: int, onset of the first code region in ms from the statt of the wav file
    :return:
    """
    return (pd.date_range(start=first_code_onset,
                          end=last_code_offset,
                          freq=f'{CODE_REGION}ms')
            .to_frame(index=False, name='code_onset')
            .assign(code_offset=lambda df: df.code_onset + pd.Timedelta(f'{CODE_REGION}ms'),
                    context_onset=lambda df: df.code_onset - pd.Timedelta(f'{CONTEXT_BEFORE}ms'),
                    context_offset=lambda df: df.code_offset + pd.Timedelta(f'{CONTEXT_AFTER}ms'),
                    since_first_code_ms=lambda df: (df.code_onset - first_code_onset).dt.total_seconds() * 1000,
                    code_onset_wav=lambda df: first_code_onset_wav + df.since_first_code_ms.astype(int))
            .drop(columns='since_first_code_ms'))


def make_intervals(sub_recordings):
    """
    Creates a population of all possible intervals to be sampled from.

    Assumptions that might become parameters later:

    - 2 minutes of buffer for context before the interval,
    - 2 minutes for the actual interval,
    - 1 minute of context after the interval,
    - start at hh:mm:00 where mm is divisible by CODE_REGION converted to minutes,
    - intervals sequence is continuous within each sub-recording and non-overlapping.

    :param sub_recordings: list of starts and ends of all sub-recordings as datetime columns `onset` and `offset` and
    an integer column `onset_wav` with the onset in ms from the start of the wav file.

    :return: pd.DataFrame with the following datetime columns: `code_onset`, `code_offset`, `context_onset`,
    and `context_offset` and an integer column `code_onset_wav` with the onset in ms from the start of the wav file.
    """
    # Find where first code region starts and last code region ends in each sub-recording
    starts_and_ends = (
        sub_recordings
        .assign(
            # Narrow the boundaries, so that there is enough space for the context.
            first_code_onset=lambda df: df.recording_start + pd.Timedelta(f'{CONTEXT_BEFORE}ms'),
            last_code_offset=lambda df: df.recording_end - pd.Timedelta(f'{CONTEXT_AFTER}ms'))
        .assign(
            # Round starts up and ends down to the closes whole number of code region durations:
            # (1:02:03, 7:45:00) -> (1:04:00, 7:44:00)
            first_code_onset=lambda df: df.first_code_onset.dt.ceil(f'{CODE_REGION}ms'),
            last_code_offset=lambda df: df.last_code_offset.dt.floor(f'{CODE_REGION}ms'),
            # Add first code region onset as ms from the wav start
            first_code_onset_recording=lambda df: (df.first_code_onset - df.recording_start).dt.total_seconds() * 1000,
            first_code_onset_wav=lambda df:
                df.recording_start_wav + df.first_code_onset_recording.astype(int)))

    # Create intervals within the boundaries we calculate above
    intervals = pd.concat(
        (_make_intervals_for_sub_recording(row.first_code_onset,
                                           row.last_code_offset,
                                           row.first_code_onset_wav)
         for _, row in starts_and_ends.iterrows()),
        ignore_index=True)

    return intervals


def add_metric(intervals, vtc_data):
    """
    For a given set of intervals calculates vtc_total_speech_duration. Works for a single recording only.

    :param intervals: a dataframe with columns `onset`, `offset`, `onset_wav` (see, e.g., `make_intervals`)
    :param vtc_data: a dataframe with VTC output for the corresponding recording
    :return: copy of intervals with a new column containing the new column 'vtc_total_speech_duration'

    Note: we already calculate energy here, now this metric, it might be a good idea to have a separate `metrics`
    module if we add more metrics.
    """
    # VTC knows nothing about no local time or sub-recordings, just the wav time. So we'll need to know when
    # intervals start and end with respect to the beginning of the wav file.
    original_columns = intervals.columns.to_list()  # we'll need them later to remove all the extra columns
    intervals = intervals.assign(
        duration_ms=lambda df: (df.code_offset - df.code_onset).dt.total_seconds() * 1000,
        code_offset_wav=lambda df: df.code_onset_wav + df.duration_ms)

    vtc_data = (vtc_data
                # We only need the speech segments
                .loc[lambda df: df.voice_type == 'SPEECH']
                # The vtc_data dataframe has 'onset' and 'duration' in seconds and these columns are of type str.
                # We'll need to parse them to floats and convert to ms.
                .assign(vtc_segment_onset_wav=lambda df: (df.onset.astype(float) * 1000).astype(int),
                        duration_ms=lambda df: (df.duration.astype(float) * 1000).astype(int),
                        # Now we can calculate the segment offset in ms
                        vtc_segment_offset_wav=lambda df: df.vtc_segment_onset_wav + df.duration_ms))

    # We'll do a full cartesian product of intervals and VTC segments and then remove pairs that don't overlap.
    return (pd.merge(intervals, vtc_data, how='cross')
            # Keep only overlapping pairs of intervals and segments
            .loc[lambda df: (df.code_onset_wav < df.vtc_segment_offset_wav)
                 & (df.code_offset_wav > df.vtc_segment_onset_wav)]
            # We only want to count the duration of the overlapping part towards the total duration in an interval.
            .assign(overlap_duration=lambda df:
                    df[['code_offset_wav', 'vtc_segment_offset_wav']].min(axis='columns')
                    - df[['code_onset_wav', 'vtc_segment_onset_wav']].max(axis='columns'))
            # Finally, we'll add up durations of the overlapping parts
            .groupby(original_columns, as_index=False)
            .overlap_duration.sum()
            .rename(columns=dict(overlap_duration='vtc_total_speech_duration'))
            # And convert original columns from a multi-index to columns
            .reset_index()
            # Covert to seconds for consistency with `blabr::get_vtc_speaker_stats`
            .assign(vtc_total_speech_duration=lambda df: df.vtc_total_speech_duration / 1000)
            )


def convert_column_to_wav_time(datetimes: pd.Series, code_onset: pd.Series, code_onset_wav: pd.Series):
    """
    Converts datetimes to time in milliseconds from the wav file start.
    :param datetimes: datetimes to convert
    :param code_onset: datetimes for which we now their wav-based version
    :param code_onset_wav: that's code_onset's wav-based version
    :return: pd.Series of integer values
    """
    ms_since_code_onset = ((datetimes - code_onset).dt.total_seconds() * 1000).astype(int)
    return code_onset_wav + ms_since_code_onset


def convert_df_to_wav_time(df):
    """
    Converts 'code_onset', 'code_offset', 'context_onset', 'context_offset' to their *_wav versions and removes them.
    :param df:
    :return:
    """
    datetime_columns = ['code_onset', 'code_offset', 'context_onset', 'context_offset']
    for column in datetime_columns:
        df[f'{column}_wav'] = convert_column_to_wav_time(df[column], df.code_onset, df.code_onset_wav)

    return df.drop(columns=datetime_columns)


def select_best_intervals(intervals, n_to_select, existing_code_intervals=None):
    """
    Select INTERVALS_FOR_ANNOTATION_COUNT intervals, potentially non-overlapping with existing intervals.
    :param intervals: a dataframe with code intervals with vtc_total_speech_duration calculated
    :param n_to_select: how many intervals should be selected
    :param existing_code_intervals: list of (onset_wav, offset_wav) that selected intervals shouldn't overlap with
    :return: (context_intervals, ranks) where
      context_intervals - list of selected intervals as (context_onset, context_offset) tuples sorted by onsets.
      ranks - list of ranks of selected intervals. If there is no overlap with existing_code_intervals, this should be a
      list of numbers from 1 to INTERVALS_FOR_ANNOTATION_COUNT in order corresponding to context_intervals.
    """
    intervals = convert_df_to_wav_time(intervals)
    metric_to_maximize = 'vtc_total_speech_duration'

    # Mark intervals that do not overlap with existing_code_intervals
    if existing_code_intervals is not None:
        is_not_overlapping = (
            intervals
            .apply(lambda row:
                   all(row.code_offset_wav <= existing_onset
                       or row.code_onset_wav >= existing_offset
                       for (existing_onset, existing_offset)
                       in existing_code_intervals),
                   axis='columns'))
    else:
        is_not_overlapping = pd.Series([True] * intervals.shape[0])

    # Check that we have enough intervals to sample from
    assert is_not_overlapping.sum() >= n_to_select

    best_intervals = (
        intervals
        # We'll rank the intervals before removing the overlapping ones so that we have information about how many we
        # had to skip.
        .assign(rank=lambda df: df[metric_to_maximize].rank(method='first', ascending=False))
        .loc[is_not_overlapping]
        .copy()
        .assign(maximized_metric=metric_to_maximize,
                metric_value=lambda df: df[metric_to_maximize],
                # rank2 will be used to select best intervals after removing overlapping ones
                rank2=lambda df: df[metric_to_maximize].rank(method='first', ascending=False))
        .loc[lambda df: df['rank2'] <= n_to_select]
        .drop(columns=[metric_to_maximize, 'rank2'])
        .reset_index(drop=True)
    )

    return best_intervals


def _extract_interval_info(eaf: EafPlus):
    """
    Extracts info from all ACLEW tiers for all intervals already in eaf.
    :param eaf: eaf to extract info from
    :return: pd.DataFrame with the following columns: code_onset, code_offset, context_onset, context_offset, code_num, on_off
    """
    code_intervals = eaf.get_time_intervals('code')
    code_onsets, code_offsets = zip(*code_intervals)

    context_intervals = eaf.get_time_intervals('context')
    context_onsets, context_offsets = zip(*context_intervals)

    code_nums = eaf.get_values('code_num')
    on_offs = eaf.get_values('on_off')
    sampling_types = eaf.get_values('sampling_type')

    return pd.DataFrame.from_dict(dict(
        sampling_type=sampling_types,
        code_onset_wav=code_onsets,
        code_offset_wav=code_offsets,
        context_onset_wav=context_onsets,
        context_offset_wav=context_offsets,
        code_num=code_nums,
        on_off=on_offs))



import pandas as pd

FIXME = '***FIX ME***'


def read_annotations_csv(path):
    """
    Reads an annotation csv file (either a csv exported from cha/opf or a sparse_code csv file) in a specific way: no
    strings are be parsed as NaN and the engine used is python. No idea why the engine bit is important.
    :param path: path to an annotation csv file
    :return: a pandas dataframe
    """
    return pd.read_csv(path, keep_default_na=False, engine='python')


def create_merged(file_new, file_old, file_merged, mode):
    """
    Merges annotations exported with export_opf_to_csv/export_cha_to_csv with the previously exported file that has
    the basic level of all words already added. Merging is done based on annotation ids (annotid) that each word in both
    files should have. For new or changed words, the basiclevel column will have ***FIX ME***
    :param file_new: path to a csv with exported annotations without basic level data
    :param file_old: path to a previous version of exported annotation with basic level data already added
    :param file_merged: path to the output file
    :param mode: 'audio'|'video' - which modality these files came from
    :return: (old_error, edit_word, new_word) - tuple of boolean values:
        old_error - were there duplicate annotids in the file with basic level data?
        edit_word - were there any changes to any of the words?
        new_word - are there new words in the exported annotations?
    """
    # print(mode)
    # print(file_old)
    # """
    if mode == "audio":
        annotid_col = "annotid"
        word_col = "word"
    elif mode == "video":
        annotid_col = "labeled_object.id"
        word_col = "labeled_object.object"
    else:
        print("Wrong mode value")
        return [], [], []
    # """

    # annotid_col = "annotid"
    # word_col = "word"
    # basic_level_col = "basic_level"

    old_error = False
    edit_word = False
    new_word = False

    old_df = read_annotations_csv(file_old)
    new_df = read_annotations_csv(file_new)

    # The basic level column in some video files is called basic_level, in others - labeled_object.basic_level. Let's
    # find which it is.
    # The code below will implicitly break if there are multiple columns whose name contains "basic_level"
    [old_basic_level_col] = old_df.columns[old_df.columns.str.contains('basic_level')]

    # For consistent naming, let's change it to 'basic_level'.
    basic_level_col = 'basic_level'
    old_df.rename(columns={old_basic_level_col: basic_level_col}, inplace=True)

    merged_df = pd.DataFrame(columns=old_df.columns.values)

    # df = df.rename(columns={'oldName1': 'newName1'})
    for index, new_row in new_df.iterrows():

        # word = ''
        to_add = new_row
        annot_id = new_row[annotid_col]
        tmp = old_df[old_df[annotid_col] == annot_id]
        # print(len(tmp.index))

        word = new_row[word_col]
        # tier = new_row['tier']
        # spk = new_row['speaker']
        # utt_type = new_row['utterance_type']
        # obj_pres = new_row['object_present']
        # ts = new_row['timestamp']

        # As far as I can tell, `while` is used here instead of `if` so that we can do `break`.
        # TODO:
        #   1. Rewrite this, so the logic is more clear: there is exactly one branch that leads to using the old word.
        #   2. Switch to using a join.
        while len(tmp.index) != 0:  # if the id already exists in the old df, check that the words/ts? do match

            if len(tmp.index) > 1:
                # One source of multiple matches are empty annotid's that comments might have
                if annot_id != '':
                    print("ERROR: annotid not unique in old version : ", annot_id)  # raise exception
                    old_error = True
                to_add[basic_level_col] = FIXME
                merged_df = merged_df.append(to_add)
                break

            old_row = tmp.iloc[0]

            # if new_row[:, new_row.columns != "basic_level"].equals(old_row[:, old_row.columns != "basic_level"]):
            if word == old_row[word_col]:
                # print("old", word)
                # check codes as well to know if something changed?
                to_add[basic_level_col] = old_row[basic_level_col]
                merged_df = merged_df.append(to_add)
                break
            else:
                # print("old but different", word)
                to_add[basic_level_col] = FIXME
                merged_df = merged_df.append(to_add)
                edit_word = True
                break

        else:  # if the id is new: no info to retrieve, add row from new
            # print(word)
            if word != '':
                # print("new", word)
                to_add[basic_level_col] = FIXME
                merged_df = merged_df.append(to_add)
                new_word = True
    # print(merged_df)
    merged_df = merged_df.loc[:, ~merged_df.columns.str.contains('^Unnamed')]

    # Remove comments from the video annotations where the comments live in separate rows (for audio, there is a
    # "comment" column)
    if mode == "video":
        is_comment = merged_df[word_col].str.startswith('%com')
        merged_df = merged_df[~is_comment]

    merged_df.to_csv(file_merged, index=False)

    return old_error, edit_word, new_word




from setuptools import setup, find_packages

setup(
    name="blabpy",
    version="0.11.5",
    packages=find_packages(),
    python_requires='>=3.7',
    install_requires=['pandas', 'numpy', 'pyarrow', 'pympi-ling', 'pydub', 'StrEnum'],
    package_data={'blabpy': ['vihi/intervals/etf_templates/*.etf',
                             'vihi/intervals/etf_templates/*.pfsx']},
    entry_points={
        'console_scripts':
            ['vihi_make_random_regions = blabpy.vihi.intervals.cli:cli_batch_create_files_with_random_regions']
    }
)



from functools import lru_cache
from pathlib import Path

from ..paths import get_pn_opus_path

AUDIO = 'Audio'
VIDEO = 'Video'
DROPPED_CHILDREN = (5, 24)
ALL_CHILDREN = tuple(child for child in range(1, 46 + 1) if child not in DROPPED_CHILDREN)
ALL_MONTHS = range(6, 17 + 1)
ANNOTATION_FILE_COUNT = 527
MISSING_AUDIO_RECORDINGS = ((22, 9),)
MISSING_VIDEO_RECORDINGS = ((17, 6),)


def ensure_folder_exists_and_empty(folder_path):
    """
    Check that folder is either empty or does not yet exist. In the latter case, creates it.
    :param folder_path:
    :return:
    """
    assert not (folder_path.exists() and any(folder_path.iterdir())), \
        'The folder should be empty or not yet exist'
    folder_path.mkdir(parents=True, exist_ok=True)


def get_seedlings_path():
    """
    Finds the path to the Seedlings folder on PN-OPUS
    :return: Path object
    """
    return get_pn_opus_path() / 'Seedlings'


def _normalize_child_month(child, month):
    """
    Converts child and month code to the two-digit (01,...,10,11,..) string representation
    :param child: int or str
    :param month: int or str
    :return: (str, str) tuple
    """
    month_str = f'{int(month):02}'
    child_str = f'{int(child):02}'
    return child_str, month_str


def _get_home_visit_folder(child, month):
    seedlings_path = get_seedlings_path()
    child, month = _normalize_child_month(child=child, month=month)
    child_month_dir = seedlings_path / 'Subject_Files' / child / f'{child}_{month}'
    return child_month_dir / 'Home_Visit'


def _get_coding_folder(child, month):
    return _get_home_visit_folder(child=child, month=month) / 'Coding'


def _get_analysis_folder(child, month):
    return _get_home_visit_folder(child=child, month=month) / 'Analysis'


def _check_modality(modality):
    assert modality in (AUDIO, VIDEO), f'Modality must be either Audio or Video but was {modality} instead'


def _get_annotation_path(child, month, modality):
    """
    Finds path to the opf/cha files
    :param modality: 'Audio'/'Video'
    :return: Path object
    """
    coding_folder = _get_coding_folder(child=child, month=month)
    child, month = _normalize_child_month(child=child, month=month)
    _check_modality(modality)
    if modality == AUDIO:
        extension = 'cha'
    elif modality == VIDEO:
        extension = 'opf'

    path = coding_folder / f'{modality}_Annotation' / f'{child}_{month}_sparse_code.{extension}'
    if not path.exists():
        raise FileNotFoundError()

    return path


def get_opf_path(child, month):
    return _get_annotation_path(child=child, month=month, modality=VIDEO)


def get_cha_path(child, month):
    return _get_annotation_path(child=child, month=month, modality=AUDIO)


def _get_all_paths(get_single_file_function, missing_child_month_combinations, **kwargs):
    """
    Runs get_single_file_function on all child-month combinations skipping files that do not exist and checking the
    total number at the end.
    :return: list of Path objects
    """
    paths = [get_single_file_function(child=child, month=month, **kwargs)
             for child in ALL_CHILDREN for month in ALL_MONTHS
             if (child, month) not in missing_child_month_combinations]
    assert len(paths) == ANNOTATION_FILE_COUNT

    return paths


@lru_cache(maxsize=None)  # do this just once
def get_all_opf_paths():
    return _get_all_paths(get_single_file_function=get_opf_path,
                          missing_child_month_combinations=MISSING_VIDEO_RECORDINGS)


@lru_cache(maxsize=None)  # do this just once
def get_all_cha_paths():
    return _get_all_paths(get_single_file_function=get_cha_path,
                          missing_child_month_combinations=MISSING_AUDIO_RECORDINGS)


def get_basic_level_path(child, month, modality):
    _check_modality(modality)
    analysis_folder = _get_analysis_folder(child=child, month=month)
    child, month = _normalize_child_month(child=child, month=month)
    path = analysis_folder / f'{modality}_Analysis' / f'{child}_{month}_{modality.lower()}_sparse_code.csv'

    if not path.exists():
        raise FileNotFoundError(path.absolute())

    return path


def _parse_out_child_and_month(file_path_or_name):
    file_name = Path(file_path_or_name).name
    child, month, *_ = file_name.split('_')
    return dict(child=int(child), month=int(month))


@lru_cache(maxsize=None)  # do this just once
def get_all_basic_level_paths(modality):
    _check_modality(modality)
    if modality == AUDIO:
        missing_child_month_combinations = MISSING_AUDIO_RECORDINGS
    elif modality == VIDEO:
        missing_child_month_combinations = MISSING_VIDEO_RECORDINGS

    return _get_all_paths(get_single_file_function=get_basic_level_path,
                          missing_child_month_combinations=missing_child_month_combinations, modality=modality)



from pkgutil import walk_packages
from importlib import import_module

from IPython import get_ipython


def autoreload_all_modules_in_package(package):
    """
    Registers all modules in packages to be autoreloaded by the autoreload IPython extension. The extension should
    already be loaded before calling this function.
    :param package: module object whose submodule and deeper nested modules should be autoreloaded.
    :return:
    """
    package_name = package.__name__
    run_magic = get_ipython().magic
    modules = [module_info.name for module_info
               in walk_packages(path=package.__path__, prefix=f'{package_name}.')]
    modules.append(package_name)
    for module in modules:
        run_magic(f"aimport {module}")


def turn_on_autoreloading():
    run_magic = get_ipython().magic
    run_magic("load_ext autoreload")
    run_magic("autoreload 1")


def autoreload_blabpy():
    blabpy = import_module('blabpy')
    turn_on_autoreloading()
    autoreload_all_modules_in_package(blabpy)



import ast
from pathlib import Path

import pytest
import pandas as pd

from blabpy.seedlings.cha import export_cha_to_csv
from blabpy.seedlings.listened_time import listen_time_stats_for_report, _get_subregion_count
from blabpy.seedlings.paths import ALL_CHILDREN, ALL_MONTHS, MISSING_AUDIO_RECORDINGS, get_cha_path
from blabpy.utils import text_file_checksum


def _possibly_interpret_as_list(possibly_list):
    if isinstance(possibly_list, str) and possibly_list.startswith('[') and possibly_list.endswith(']'):
        return ast.literal_eval(possibly_list)
    else:
        return possibly_list


def _get_test_cha_path(child, month):
    return Path(f'data/annotated_cha/annotated_cha/{child:02}_{month:02}_sparse_code.cha')


@pytest.fixture(scope='module')
def listen_time_stats_df():
    return pd.read_csv('data/listen_time_stats.csv')


@pytest.mark.parametrize(argnames=('child', 'month'),
                         argvalues=[(child, month)
                                    for child in ALL_CHILDREN for month in ALL_MONTHS
                                    if (child, month) not in MISSING_AUDIO_RECORDINGS])
def test_listen_time_stats_for_report(listen_time_stats_df, child, month):
    subregion_count = _get_subregion_count(child=child, month=month)
    cha_path = _get_test_cha_path(child, month)
    stats_correct = listen_time_stats_df.set_index('filename').loc[cha_path.name].to_dict()

    stats = listen_time_stats_for_report(clan_file_text=cha_path.read_text(), subregion_count=subregion_count)

    for key, correct_value in stats_correct.items():
        correct_value = _possibly_interpret_as_list(correct_value)
        assert stats[key] == correct_value


@pytest.fixture(scope='module')
def cha_export_checksums_df():
    return pd.read_csv('data/cha_parsing_checksums.csv').set_index('cha_filename')


@pytest.mark.parametrize(argnames=('child', 'month'),
                         argvalues=[(child, month)
                                    for child in ALL_CHILDREN for month in ALL_MONTHS
                                    if (child, month) not in MISSING_AUDIO_RECORDINGS])
def test_export_cha_to_csv(cha_export_checksums_df, child, month, tmpdir):
    cha_path = _get_test_cha_path(child, month)
    cha_checksum = text_file_checksum(cha_path)
    export_cha_to_csv(cha_path, tmpdir)
    exported_csv_path = tmpdir / cha_path.name.replace(".cha", "_processed.csv")
    exported_csv_checksum = text_file_checksum(exported_csv_path)

    checksums = cha_export_checksums_df.loc[cha_path.name]

    assert cha_checksum == checksums.cha_checksum
    assert exported_csv_path.exists()
    assert exported_csv_checksum == checksums.exported_csv_checksum
